"""
Classes for Gaussian Process Regression fitting of 1D fusion profile data with errorbars.

Requires installation of the GPR1D package, now available on PyPI.

These classes were developed by Aaron Ho.
"""

import numpy as np
import GPR1D

__all__ = ['RationalQuadraticFGPR1D', 'GibbsInverseGaussianFGPR1D', 'Matern52FGPR1D']


class RationalQuadraticFGPR1D(GPR1D.SimplifiedGaussianProcessRegression1D):
    """
    Implements a GPR with a simple rational quadratic
    kernel, including heteroscedastic error fitting, ie.
    the error is allowed to vary in x. This is the most
    robust kernel, generally able to fit most profiles
    that do not have sudden dramatic changes in gradient,
    such as pedestals.

    Optimization of hyperparameters is only performed
    *once* using settings at the time of the *first*
    call! All subsequent calls use the results of the
    first optimization.
    """
    def __init__(self,x,y,ye,xe=None,amplitude=0.5,length_scale=0.3,regularization_parameter=1.0,epsilon=1.0e-2,nrestarts=5):
        """
        Defines customized GaussianProcessRegression1D instance with
        a pre-defined rational quadratic kernel for both data fit and
        error fit. Input parameters reduced only to essentials and
        most crucial knobs for fine-tuning.

        :arg x: array. Vector of x-values corresponding to data to be fitted.

        :arg y: array. Vector of y-values corresponding to data to be fitted.

        :arg ye: array. Vector of y-errors corresponding to data to be fitted.

        :kwarg xe: array. Optional vector of x-errors corresponding to data to be fitted, setting to anything other than None toggles x-error considerations.

        :kwarg amplitude: float. Hyperparameter related to allowable variation in fitted value, smaller value enforces flatter profiles.

        :kwarg length_scale: float. Hyperparameter related to correlation length between points in profile, larger value makes profiles flatter.

        :kwarg regularization_parameter: float. Parameter adjusting penalty on kernel complexity, larger value makes profiles smoother.

        :kwarg epsilon: float. Convergence criterion on change in log-marginal-likelihood, relaxing this parameter speeds up optimization at cost of goodness-of-fit.

        :kwarg nrestarts: int. Number of kernel restarts, larger number improves robustness of algorithm at the cost of speed.

        :returns: none.
        """
        self._amplitude_initial = amplitude
        self._length_scale_initial = length_scale
        kernel = GPR1D.RQ_Kernel(1.0e0,self._length_scale_initial,1.0e1)
        kernel_hyppar_bounds = np.atleast_2d([[self._amplitude_initial * 0.5,self._length_scale_initial * 0.5,5.0e0], \
                                              [self._amplitude_initial * 2.0,self._length_scale_initial * 2.0,2.0e1]])
        optimizer_gain = 1.0e-1
        super(RationalQuadraticFGPR1D,self).__init__(kernel,x,y,ye,xe,kernel_hyppar_bounds,regularization_parameter,epsilon,nrestarts)

        # Overwrite the preset kernel optimizer settings
        self.set_search_parameters(epsilon=epsilon,method='adam',spars=[optimizer_gain,0.4,0.8])

        # Define a zero derivative constraint at the magnetic axis
        self.set_raw_data(dxdata=[0.0],dydata=[0.0],dyerr=[0.0])

        # Overwrite the error kernel preset by the simplified GPR1D class
        error_kernel = GPR1D.RQ_Kernel(3.0e-1,4.0e-1,4.0e1)
        error_kernel_hyppar_bounds = np.atleast_2d([[1.0e-1,2.0e-1,3.0e1],[5.0e-1,5.0e-1,5.0e1]])
        self.set_error_kernel(kernel=error_kernel,kbounds=error_kernel_hyppar_bounds,regpar=5.0,nrestarts=0)
        self.set_error_search_parameters(epsilon=1.0e-1,method='adam',spars=[optimizer_gain,0.4,0.8])


class GibbsInverseGaussianFGPR1D(GPR1D.SimplifiedGaussianProcessRegression1D):
    """
    Implements a GPR with a Gibbs kernel, augmented by
    an inverse Gaussian warping function, including
    heteroscedastic error fitting, ie. the error is
    allowed to vary in x. This kernel is best suited
    to density pedestal fitting, provided the peak
    location parameter is set approximately equal to
    the x-value of the sharpest gradient in the pedestal.

    Note that the peak location is NOT a hyperparameter,
    ie. the optimization routine does not change this
    value!

    Optimization of hyperparameters is only performed
    *once* using settings at the time of the *first*
    call! All subsequent calls use the results of the
    first optimization.
    """
    def __init__(self,x,y,ye,xe=None,amplitude=0.5,length_scale=0.2,peak_width=0.15,peak_location=1.0,regularization_parameter=1.0,epsilon=1.0e-2,nrestarts=5):
        """
        Defines customized GaussianProcessRegression1D instance with
        a pre-defined Gibbs kernel with an inverse Gaussian warping
        function for the data fit and a simple rational quadratic
        kernel for the error fit. Input parameters reduced only to
        essentials and most crucial knobs for fine-tuning.

        :arg x: array. Vector of x-values corresponding to data to be fitted.

        :arg y: array. Vector of y-values corresponding to data to be fitted.

        :arg ye: array. Vector of y-errors corresponding to data to be fitted.

        :kwarg xe: array. Optional vector of x-errors corresponding to data to be fitted, setting to anything other than None toggles x-error considerations.

        :kwarg amplitude: float. Hyperparameter related to allowable variation in fitted value, smaller value enforces flatter profiles.

        :kwarg length_scale: float. Hyperparameter related to correlation length between points in profile, larger value makes profiles flatter.

        :kwarg peak_width: float. Hyperparameter related to minimum correlation length, smaller value captures pedestal better. A hard-coded limit of 0.8 * length_scale is present to provide stability.

        :kwarg peak_location: float. Parameter specifying the location of the minimum correlation length, recommended to set slightly outside the pedestal.

        :kwarg regularization_parameter: float. Parameter adjusting penalty on kernel complexity, larger value makes profiles smoother.

        :kwarg epsilon: float. Convergence criterion on change in log-marginal-likelihood, relaxing this parameter speeds up optimization at cost of goodness-of-fit.

        :kwarg nrestarts: int. Number of kernel restarts, larger number improves robustness of algorithm at the cost of speed.

        :returns: none.
        """
        self._amplitude_initial = amplitude
        self._length_scale_initial = length_scale
        self._peak_width_initial = peak_width
        self._peak_location = peak_location
        warpfunc = GPR1D.IG_WarpingFunction(self._length_scale_initial,self._length_scale_initial * 0.6,self._peak_width_initial,self._peak_location,0.8)
        kernel = GPR1D.Gibbs_Kernel(self._amplitude_initial,warpfunc)
        kernel_hyppar_bounds = np.atleast_2d([[self._amplitude_initial * 0.5,self._length_scale_initial * 0.5,self._length_scale_initial * 0.6 * 0.8,self._peak_width_initial * 0.8], \
                                              [self._amplitude_initial * 2.0,self._length_scale_initial * 2.0,self._length_scale_initial * 0.6 * 1.2,self._peak_width_initial * 1.2]])
        optimizer_gain = 1.0e-1
        super(GibbsInverseGaussianFGPR1D,self).__init__(kernel,x,y,ye,xe,kernel_hyppar_bounds,regularization_parameter,epsilon,nrestarts)

        # Overwrite the preset kernel optimizer settings
        self.set_search_parameters(epsilon=epsilon,method='adam',spars=[optimizer_gain,0.4,0.8])

        # Define a zero derivative constraint at the magnetic axis
        self.set_raw_data(dxdata=[0.0],dydata=[0.0],dyerr=[0.0])

        # Overwrite the error kernel preset by the simplified GPR1D class
        error_kernel = GPR1D.RQ_Kernel(3.0e-1,4.0e-1,4.0e1)
        error_kernel_hyppar_bounds = np.atleast_2d([[1.0e-1,2.0e-1,3.0e1],[5.0e-1,5.0e-1,5.0e1]])
        self.set_error_kernel(kernel=error_kernel,kbounds=error_kernel_hyppar_bounds,regpar=5.0,nrestarts=0)
        self.set_error_search_parameters(epsilon=1.0e-1,method='adam',spars=[optimizer_gain,0.4,0.8])


class Matern52FGPR1D(GPR1D.SimplifiedGaussianProcessRegression1D):
    """
    Implements a GPR with a Matern kernel with an order
    parameter of 5/2, including heteroscedastic error
    fitting, ie. the error is allowed to vary in x. This
    kernel retains the best feature resolution in
    general, and is best suited to temperature profiles.

    .. note::

        This can suffer from extreme overfitting due
        to existence of many local minima in hyperparameter
        space. This means the selection of an appropriate
        initial length scale is more important with this
        kernel option.

    Optimization of hyperparameters is only performed
    *once* using settings at the time of the *first*
    call! All subsequent calls use the results of the
    first optimization.
    """
    def __init__(self,x,y,ye,xe=None,amplitude=0.5,length_scale=1.0,regularization_parameter=1.0,epsilon=1.0e-2,nrestarts=5):
        """
        Defines customized GaussianProcessRegression1D instance with
        a pre-defined Matern half-integer kernel for the data fit
        and a simple rational quadratic kernel for the error fit.
        Input parameters reduced only to essentials and most crucial
        knobs for fine-tuning.

        :arg x: array. Vector of x-values corresponding to data to be fitted.

        :arg y: array. Vector of y-values corresponding to data to be fitted.

        :arg ye: array. Vector of y-errors corresponding to data to be fitted.

        :kwarg xe: array. Optional vector of x-errors corresponding to data to be fitted, setting to anything other than None toggles x-error considerations.

        :kwarg amplitude: float. Hyperparameter related to allowable variation in fitted value, smaller value enforces flatter profiles.

        :kwarg length_scale: float. Hyperparameter related to correlation length between points in profile, larger value makes profiles flatter.

        :kwarg regularization_parameter: float. Parameter adjusting penalty on kernel complexity, larger value makes profiles smoother.

        :kwarg epsilon: float. Convergence criterion on change in log-marginal-likelihood, relaxing this parameter speeds up optimization at cost of goodness-of-fit.

        :kwarg nrestarts: int. Number of kernel restarts, larger number improves robustness of algorithm at the cost of speed.

        :returns: none.
        """
        self._amplitude_initial = amplitude
        self._length_scale_initial = length_scale
        kernel = GPR1D.Matern_HI_Kernel(self._amplitude_initial,self._length_scale_initial,2.5)
        kernel_hyppar_bounds = np.atleast_2d([[self._amplitude_initial * 0.5,self._length_scale_initial * 0.8], \
                                              [self._amplitude_initial * 2.0,self._length_scale_initial * 1.2]])
        optimizer_gain = 1.0e-1
        super(Matern52FGPR1D,self).__init__(kernel,x,y,ye,xe,kernel_hyppar_bounds,regularization_parameter,epsilon,nrestarts)

        # Overwrite the preset kernel optimizer settings
        self.set_search_parameters(epsilon=epsilon,method='adam',spars=[optimizer_gain,0.4,0.8])

        # Define a zero derivative constraint at the magnetic axis
        self.set_raw_data(dxdata=[0.0],dydata=[0.0],dyerr=[0.0])

        # Overwrite the error kernel preset by the simplified GPR1D class
        error_kernel = GPR1D.RQ_Kernel(3.0e-1,4.0e-1,4.0e1)
        error_kernel_hyppar_bounds = np.atleast_2d([[1.0e-1,2.0e-1,3.0e1],[5.0e-1,5.0e-1,5.0e1]])
        self.set_error_kernel(kernel=error_kernel,kbounds=error_kernel_hyppar_bounds,regpar=5.0,nrestarts=0)
        self.set_error_search_parameters(epsilon=1.0e-1,method='adam',spars=[optimizer_gain,0.4,0.8])



#!/usr/bin/env python

import os
import sys
import re
import copy
import numpy as np

import gpr1dfusion      # Required import, only works after using 'pip install'


def run_test():
    """
    A demonstration script for the classes within the gpr1dfusion module.

    Due to the iterative nature of the optimization method and the
    random nature of the kernel restart function, the results may
    not be exactly the same for consecutive runs of this demo
    script. However, all fits should fall within the fit error
    ranges of each other, unless the optimization algorithm has
    not converged.
    """

    ### Some basic setup

    plot_save_directory = './gpr1dfusion_test_plots/'
    if not plot_save_directory.endswith('/'):
        plot_save_directory = plot_save_directory + '/'
    if not os.path.isdir(plot_save_directory):
        os.makedirs(plot_save_directory)


    ### Generating sample data

    # Make basic function data
    x_spread = 0.01
    y_spread = 0.25
    intercept = 3.0
    slope1 = 1.0
    x_values = np.linspace(0.0,1.0,21)
    y_values = slope1 * x_values + intercept
    boundary1 = 0.3
    slope2 = 16.0
    boundary1_filter = (x_values >= boundary1)
    y_values[boundary1_filter] = y_values[boundary1_filter] - slope2 * (x_values[boundary1_filter] - boundary1)
    boundary2 = 0.7
    boundary2_filter = (x_values >= boundary2)
    y_values[boundary2_filter] = y_values[boundary2_filter] + (slope2 + slope1) * (x_values[boundary2_filter] - boundary2)

    # Add random error to generated data points
    raw_x_values = x_values + x_spread * np.random.randn(x_values.size)
    raw_y_values = y_values + y_spread * np.random.randn(y_values.size)
    raw_x_errors = np.full(raw_x_values.shape,x_spread)
    raw_y_errors = np.full(raw_y_values.shape,y_spread)
    raw_intercept = raw_y_values[0]



    ### Fitting

    fit_x_values = np.linspace(0.0,1.0,100)

    # Define a kernel to fit the data itself
    #     Rational quadratic kernel is usually robust enough for general fitting
    gpr_object = gpr1dfusion.RationalQuadraticFGPR1D(raw_x_values,raw_y_values,raw_y_errors,amplitude=1.0e0,length_scale=1.0e0,nrestarts=5)

    #     Perform the fit with kernel restarts, returns fit results
    (fit_y_values,fit_y_errors,fit_dydx_values,fit_dydx_errors) = gpr_object(fit_x_values)

    #     Grab optimized kernel settings - easy way to minimize data storage requirements for fit reproduction
    (gp_kernel_name,gp_kernel_hyppars,gp_fit_regpar) = gpr_object.get_gp_kernel_details()

    #     Grab the log-marginal-likelihood of fit
    fit_lml = gpr_object.get_gp_lml()



    ### Sampling distribution

    num_samples = 10
    sample_array = None
    dsample_array = None

    # Samples the fit distribution
    for ii in np.arange(0,num_samples):
        sample = gpr_object.sample(fit_x_values,derivative=False)
        sample_array = np.vstack((sample_array,sample)) if sample_array is not None else copy.deepcopy(sample)

    # Samples the derivative distribution
    for ii in np.arange(0,num_samples):
        dsample = gpr_object.sample(fit_x_values,derivative=True)
        dsample_array = np.vstack((dsample_array,dsample)) if dsample_array is not None else copy.deepcopy(dsample)

    # Computing statistics of sampled profiles
    sample_mean = np.nanmean(sample_array,axis=0)
    dsample_mean = np.nanmean(dsample_array,axis=0)
    sample_std = np.nanstd(sample_array,axis=0)
    dsample_std = np.nanstd(dsample_array,axis=0)



    ### Printing

    gp_str = "\n--- GPR Fit ---\n\n"
    gp_str = gp_str + "Kernel name: %30s\n" % (gp_kernel_name)
    gp_str = gp_str + "Regularization parameter: %17.4f\n" % (gp_fit_regpar)
    gp_str = gp_str + "Optimized kernel hyperparameters:\n"
    for hh in np.arange(0,gp_kernel_hyppars.size):
        gp_str = gp_str + "%15.6e" % (gp_kernel_hyppars[hh])
    gp_str = gp_str + "\n\n"
    gp_str = gp_str + "Log-marginal-likelihood: %18.6f\n" % (fit_lml)

    print(gp_str)



    ### Plotting

    plt = None
    try:
        import matplotlib.pyplot as plt
    except:
        plt = None

    if plt is not None:

        plot_sigma = 1.0

        # Raw data with GPR fit and error, only accounting for y-errors
        plot_raw_y_errors = plot_sigma * raw_y_errors
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.errorbar(raw_x_values,raw_y_values,yerr=plot_raw_y_errors,ls='',marker='.',color='g')
        ax.plot(fit_x_values,fit_y_values,color='r')
        plot_fit_y_lower = fit_y_values - plot_sigma * fit_y_errors
        plot_fit_y_upper = fit_y_values + plot_sigma * fit_y_errors
        ax.fill_between(fit_x_values,plot_fit_y_lower,plot_fit_y_upper,facecolor='r',edgecolor='None',alpha=0.2)
        for ii in np.arange(0,num_samples):
            ax.plot(fit_x_values,sample_array[ii,:],color='k',alpha=0.5)
        ax.set_xlim(0.0,1.0)
        fig.savefig(plot_save_directory+'rq_test.png')
        plt.close(fig)

        # Derivative of GPR fit and error, only accounting for y-errors
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.plot(fit_x_values,fit_dydx_values,color='r')
        plot_fit_dydx_lower = fit_dydx_values - plot_sigma * fit_dydx_errors
        plot_fit_dydx_upper = fit_dydx_values + plot_sigma * fit_dydx_errors
        ax.fill_between(fit_x_values,plot_fit_dydx_lower,plot_fit_dydx_upper,facecolor='r',edgecolor='None',alpha=0.2)
        for ii in np.arange(0,num_samples):
            ax.plot(fit_x_values,dsample_array[ii,:],color='k',alpha=0.5)
        ax.set_xlim(0.0,1.0)
        fig.savefig(plot_save_directory+'rq_dtest.png')
        plt.close(fig)

        print("Results of RQ kernel test plotted in directory %s\n" % (plot_save_directory))

    else:

        print("   Module matplotlib not found. Skipping plotting of demonstration results.\n")

    print("Demonstration script successfully completed!\n")


def main():

    run_test()


if __name__ == "__main__":

    main()
